import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.security.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

/**
    AutoGrader - a program that automatically grades programming assignments.
    Copyright (C) 2006  Pete Nordquist   (nordquip at sou.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
	(See the file COPYING)


	This program uses code that is Copyright (c) 2005 - 2006 Swiss Federal
	Institute of Technology (ETH Zurich),
	Department of Computer Science (http://www.inf.ethz.ch),
	Christian Plattner. All rights reserved.
	(See the file LICENSE.txt)

 * Writes a test harness to the client.
 * Compiles the test harness and client code.
 * Executes the resulting program to execute tests of the client's code and report
 * grades.
 * Program is passed:
    client home directory,
    file name where grades are written
    grade file format is:
        <pointsSubtracted><delim> Test description: <PASSED or FAILED> - <failure message>
        -0<delim> Test 1 description: PASSED
        -10<delim> Test 2 description: FAILED - <failure message>
			(char) 0, now used for the delimiter instead of just ',', so ',' can be used in 
			messages and descriptions
            They can read PUT.java to see the test cases.
 *  LateFeeTable format
 *  <points_possible_for_the_lab>(char) 0,  // must be line 1 (char) 0, required
 *  -<points_to_subtract_after_date>(char) 0,<date>
 *  ...
 *  For example:
 *  102,Points Possible with extra credit
 *  -5<delim>2006.10.07:0600 // 6AM on 20061007 subtract 5 if later than this
 *  -5<delim>2006.10.14:0600 // 6AM on 20061014 subtract 5 more if later than this
 * Copies results back to server
 *  Server Directory structure assumed to be:
 *  F:/AG/cs257/.userIDList.txt
 *  F:/AG/cs257/.labNoList.txt
 *  F:/AG/cs257/guiListFiles.zip
 *  F:/AG/cs257/<labNo>/PUT.zip  // contains all PUT source and LateFeeTable
 *  F:/AG/cs257/<labNo>/KEY.zip
 *  F:/AG/cs257/<labNo>/grades/<userID>.<labNo>.zip
 *  F:/AG/cs257/<labNo>/grades/<userID>.<labNo>.grade.txt
 *  F:/AG/cs257/keyReceived/<userID>.<labNo>.keyRecd // file contains the date the
 *     student received the key
 *  archive the whole cs257 directory as f:\AG\200601\cs257

 * Every run of the program overwrites results of any previous run.
 *
 * Reflection for compilation and invocation of compiled program from:
 * http://www.javaworld.com/javatips/jw-javatip131.html#resources
 * by Shawn Silverman
 * @author Pete Nordquist
 * @version 060908 -- Initial attempt from Shawn's code
 * @version 060915 -- compiles and executes arbitrary code on the client's machine
 * @version 070927 -- FTP to dedicated autograder account
 * @version 071001 -- Caught FTP exceptions and closed connections.
 * @version 071002 -- Now uses SCP for file transfers
 */
public class AG
{
    // constants
	private static final String delim = ((char) 0) + ",";
    // tools.jar name
    private static final String TOOLSJAR = "tools.jar";
    // client file names
    // data files
    // all necessary files, including all test harness source and the LateFeeTable
    // are zipped up on the server and transferred in a single send.
    private static final String HARNESSZIPFILENAME = "PUT.zip";
    private static final String PUTSRCFILENAME = "PUT.java";
    private static final String PUTCLASSNAME = "PUT";
    private static final String LATEFEETABLENAME = "LateFeeTable.txt";
    // result files
    // file written by PUT.java
    private static final String GRADEFILEEXT = ".grade.txt";
    // calculated grade in format to send to server
    private static final String LOGFILENAME = "log.txt";

    // server file names
    // gui lists -- cached on client for offline execution
    // userIDList is cached in the javaws working directory for security
    private static final String USERIDLISTFILENAME = ".userIDList.txt";
    private static final String LABNOLISTFILENAME = ".labNoList.txt";
    // grade collecting directory
    private static final String GRADEDIRNAME = "grades/";
    // lab key files
    // key name
    private static final String KEYFILENAME = "KEY.zip";
    // key sent indicator
    private static final String KEYRECDEXT = ".keyRecd";
    // directory on server that collects keys
    private static final String KEYRECDDIR = "keysReceived/";

    // result files
    private static final String ZIPEXT = ".zip";
    // date format string
    private static final String DATEFORMATSTRING = "yyyy.MM.dd:HHmm";

    // variables
    // the GUI
    private static AGGUI gui;
    // grade received
    private static double grade;
    // status line
    private static String stat = "";
    // inputs:
    // user's ID
    private static String userID;
    // number of the course
    private static String courseNo;
    // number of the lab
    private static String labNo;
    // root directory of source on client machine
    private static String userDir;
    // full path to the tools.jar file
    private static String toolsJarPath;

    // output stream for stdio redirection
    private static ByteArrayOutputStream baos;

	// server connection flag
	private static boolean usingLocalTestHarness;

    /**
      * Main entry point.
      */
    public static void main(String[] args)
    {
        // Create the GUI
        gui = new AGGUI();
    }

    /**
     * Determine if the paths the user gave are legitimate
     * Print errors if they are not
     * @return true if the paths the user gave exist
     */
    public static boolean checkAndSaveGUIPaths()
    {
        boolean retVal = true;
        // get the userDir
        userDir = gui.getUserDir();
        // if it does not exist, print an error message
        if (! (new File(userDir)).exists()) {
            showMsg("?CANNOT FIND DIRECTORY CONTAINING YOUR FILES.\n");
            retVal = false;
        }
        // get toolsJarPath
        toolsJarPath = gui.getToolsJarPath();
        // check this path
        File tjFile = new File(toolsJarPath);
        // if it is a directory
        if (tjFile.isDirectory()) {
            // if there is a file tools.jar in this dir
            File tjNew = new File(toolsJarPath + File.separatorChar + TOOLSJAR);
            if (tjNew.isFile()) {
                // let tjFile be this file
                tjFile = tjNew;
                // reset toolsJarPath
                toolsJarPath = tjFile.getPath();
            }
        }
        // if tjFile is a file and its name is tools.jar, it is OK
        if (! (tjFile.isFile() && tjFile.getName().equals(TOOLSJAR))) {
            showMsg("?PATH TO tools.jar IS INCORRRECT.\n");
            retVal = false;
        }
        if (retVal) {
            // add a line showing the working directory
            showMsg("WORKING IN DIRECTORY: " + userDir + "\n");
            showMsg("PATH TO COMPILER TOOLS: " + toolsJarPath + "\n");
        }
        return retVal;
    }
     
    /**
     * Get the userID and labNo lists from the server.  If can't reach the server try
     * to get the lists from the local directory.
     * Save the userID list in the javaws working directory.
     * Initialize the userDir, courseNo globals
     *
     * @return an Object array as follows:
     * a boolean - false if unable to come up with data either from the server or local
     * a String[] - the labNoList
     */
    public static Object[] populateGUILists()
    {
        // return 3 values
        Object[] retarr = new Object[3];
        boolean retVal = false; // assume the worst
        redirectStdIO();
        courseNo = gui.getCourseNo();
        String[] userIDList = null;
        String[] labNoList = null;
        String s;
        try {
            // Get Lists for gui list boxes
            // userID
            // wanted to put userID list in java's working directory,
            // not in userDir for security, but it ended up on the desktop when
            // I tested with Firefox in the lab
            showMsg("RECEIVING GUI LIST FILES ...\n");
			String tmpDirStr = System.getProperty("java.io.tmpdir");

			// connect to server
			SCP scp = new SCP();
			// ID list
            scp.transfer(
				SCP.GET,
				courseNo + "/" + USERIDLISTFILENAME,
				tmpDirStr);
			// Lab no list
            scp.transfer(
				SCP.GET,
				courseNo + "/" + LABNOLISTFILENAME,
				userDir + File.separator);
			// close connection
			scp.close();

			// read ID list into a string
			s = readFile( tmpDirStr + USERIDLISTFILENAME);
            // init local list variables from text transferred down
            userIDList = s.split("\n");

            // labNo
			s = readFile( tmpDirStr + LABNOLISTFILENAME);
            labNoList = s.split("\n");

            // got em
            retVal = true;
        } catch (Exception e) {
            try {
                // look for local copies
                s = readFile( 
                        System.getProperty("java.io.tmpdir") +
						File.separatorChar +
                        USERIDLISTFILENAME);
                userIDList = s.split("\n");
                s = readFile( getClientFileFullPath(LABNOLISTFILENAME));
                labNoList = s.split("\n");
                // got em
                retVal = true;
            } catch (Exception ex) {
                showMsg("\nCOULD NOT GET LIST FILES FROM SERVER " +
                    "and there are no local copies.\n");
                ex.printStackTrace();
                System.err.println();
                e.printStackTrace();
            }
        } finally {
            showMsg(baos.toString());
            retarr[0] = userIDList;
            retarr[1] = labNoList;
            retarr[2] = new Boolean(retVal);
            return retarr;
        }
    }

    /**
     * All actions when user submits data for processing
     */
    public static void doSubmit()
    {
        // Initialize data from GUI
        // already have userDir
        userID = gui.getUserID();
        labNo = gui.getLabNo();

		// set local test harness flag
		usingLocalTestHarness = false;
        // if couldn't get files from server:
        if (! getServerDataFiles())
            // if wasn't successful, let user know
            showMsg("\nCOULD NOT GET TEST HARNESS FILES FROM SERVER " +
                "and there are no usable local copies.\n");

        else if (! doRun())  // if does not compile and run
            showMsg("\nINVALID run -- results will not be sent.\n");

        else {
            // calculate the grade and post to the gui and gradefile
            calculateAndPostGrade();
            // send results as long as they haven't yet received the key
			if ( usingLocalTestHarness )
                showMsg("\nUSED LOCAL COPY OF TEST HARNESS -- " +
					"RESULTS WILL NOT BE SUBMITTED.\n");
            else if (! keyHasNotBeenReceived())
                showMsg("\nRESULTS WILL NOT BE SUBMITTED.\n");
            else
                sendResults();
        }
        showMsg("\nALL DONE.\n");
    }

    /**
     * get the key from the server
     * on the server, mark that the key has been received
     */
    public static void receiveKey()
    {
        // Initialize data from GUI
        // already have userDir
        userID = gui.getUserID();
        labNo = gui.getLabNo();
        try {
            // attempt to recieve key
            showMsg("RECEIVING LAB KEY ...\n");
            redirectStdIO();
            // Get Key
			// connect to server
			SCP scp = new SCP();
			// get key file
            scp.transfer(
				SCP.GET,
				courseNo + "/" + labNo + "/" + KEYFILENAME,
				userDir + File.separator);

            // get current date/time
            Date now = new Date();
            // format it 
            String nowString =
				(new SimpleDateFormat(DATEFORMATSTRING)).format(now,
					new StringBuffer(), new FieldPosition(0)).toString();
            // write it as a file to userDir
            writeFile( getClientFileFullPath(keyReceivedFileName()), nowString);
            // put local file to server
            scp.transfer(
				SCP.PUT,
				getClientFileFullPath(keyReceivedFileName()),
                courseNo + "/" + KEYRECDDIR);
			// close connection
			scp.close();

            // record in log that key was received now
            showMsg("LAB KEY RECEIVED -- see " + KEYFILENAME + "\n");

        } catch (Exception ex) { // something went wrong
            // if filenotfound
            if (ex.getCause().toString().indexOf( KEYFILENAME +
				": No such file") > -1) 
                showMsg("LAB KEY NOT YET POSTED -- please try again later.\n");
            else {
                // roll back the entire transaction
                showMsg("\nCOULD NOT GET LAB KEY FROM SERVER:\n");
                // get the stack trace
                ex.printStackTrace();
                // delete the key on the client
                (new File(getClientFileFullPath(KEYFILENAME))).delete();
            }
        } finally {
            // update the log
            showMsg(baos.toString());
        }
    }


    /**
     * ask the gui to remove the fields and leave only the display area
     */
    public static void quit()
    {
        gui.removeAllButLog();
    }

    /****************************
    P R I V A T E   M E T H O D S
    ****************************/

    /**
     * @return the name of the file that indicates that the key has been received
     */
    private static String keyReceivedFileName()
    {
        return userID + "." + labNo + KEYRECDEXT;
    }

    /**
     * @return true if and only if can contact the server and successfully confirm
     * that the key has never been received
     * false otherwise
     */
    private static boolean keyHasNotBeenReceived()
    {
        boolean retVal = false; // assume key has been received
        showMsg("CONTACTING SERVER to see if key has been previously received ...\n");
		String krfn = keyReceivedFileName();
        try {
            redirectStdIO();
            // receive key rec'd indicator file
            // read the file into a string
            getFromServer(
                courseNo + "/" + KEYRECDDIR + krfn,
				userDir + File.separator);
            String whenString = readFile( getClientFileFullPath(krfn));
            // record in log that key has been received
            showMsg("LAB KEY PREVIOUSLY RECEIVED -- " + whenString + "\n");

        } catch (Exception ex) { 
            // if the file really wasn't there, i.e., if exception cause
			// was no such file
            if (ex.getCause().toString().indexOf(krfn+": No such file") > -1) {
                showMsg("LAB KEY not yet received.  OK to submit results.\n");
                retVal = true;
            }
            // any other exception means something in the contact went wrong
            else { 
                showMsg("UNABLE TO CONTACT SERVER.\n");
            }
        } finally {
            // update the log
            showMsg(baos.toString());
            return retVal;
        }
    }

    /**
     * Compiles and runs the program under test
     */
    private static boolean doRun()
    {
        boolean retVal = false;
        try {


            // Redirect std out
            redirectStdIO();

            // remove all .class files from userDir
            File[] fa = getFilesInDirEndingWith(userDir, ".class");
            for (int i = 0; i < fa.length; i++)
                fa[i].delete();

            // get the .java files in userDir
            fa = getFilesInDirEndingWith(userDir, ".java");

            // create arguments array
            String[] compileArgs = new String[5 + fa.length];

            // set initial args
            compileArgs[0] = "-g";
            compileArgs[1] = "-classpath";
            compileArgs[2] = userDir;
            compileArgs[3] = "-d";
            compileArgs[4] = userDir;

            // build up a string for printing
            String compileArgsToString = "";
            for (int i = 0; i < 5; i++)
                compileArgsToString += compileArgs[i] + " ";
            // add on java files to compile
            for (int i = 0; i < fa.length; i++) {
                compileArgs[i+5] = fa[i].getPath();
                compileArgsToString +=
                    compileArgs[i+5].substring((userDir + File.separator).length()) + " ";
            }

            showMsg("STARTING COMPILATION ...\n");
            showMsg("javac " + compileArgsToString + "\n");

            // Create a File object on tools.jar
            File file = new File(toolsJarPath);

            // Convert File to a URL
            URL url = file.toURL();          // file:/c:/myclasses/
            URL[] urls = new URL[]{url};
            // disable the security manager
            // http://forum.java.sun.com/thread.jspa?forumID=38&threadID=706976
            System.setSecurityManager(null);
            // Create a new class loader for the url array
            ClassLoader cl = new URLClassLoader(urls);
            // Create the compiler object and do the compilation
            Class clazz = cl.loadClass("com.sun.tools.javac.Main");
            
            Method main = clazz.getMethod("compile", new Class[] { String[].class });
            int status = ((Integer)
                main.invoke(null, new Object[] { compileArgs })).intValue();
            // com.sun.tools.javac.Main javac = new com.sun.tools.javac.Main();
            // int status = javac.compile(compileArgs);

            // display compiler messages
            showMsg(baos.toString());

            // display compile status
            showMsg("\nCompilation " + 
                ((status == 0) ? "SUCCEEDED"
                    : "FAILED -- please correct errors"
                    ) +
                "\n\n");

            // if compiled OK, then run it
            if (status == 0) {
                redirectStdIO();
                // call the entry point for the program under test

                // Create a File object on the root of the directory containing the class file
                file = new File(userDir + File.separatorChar);

                // Convert File to a URL
                url = file.toURL();          // file:/c:/myclasses/
                urls = new URL[]{url};
                // Create a new class loader for the url array
                cl = new URLClassLoader(urls);

                // load in the PUT class
                clazz = cl.loadClass(PUTCLASSNAME);

                // get a method named main with a single arg of type String[]
                main = clazz.getMethod("main", new Class[] {String[].class});
                showMsg("STARTING PROGRAM EXECUTION ...\n");
                // start main -- pass in the grade file name
                main.invoke(null, new Object[] { new String[]{clientGradeFileFullPath()}});
                // made it
                retVal = true;
            }
        } catch (Exception ex) {
            // Exception in the main method that we just tried to run
            ex.printStackTrace();
        } finally {
            // Print output from PUT
            showMsg(baos.toString());
            return retVal;
        }
    }
    
    /**
     * calculate the grade and post to the gui and gradefile
     */
    private static void calculateAndPostGrade() 
    {
        try {
            String splitRegEx = "\n|" + ((char) 0) + ",";
            String formatString = "%6.2f   %-60s%n";
            // declare accumulator
            double total = 0.0;
            double d;
            // Create StringBuilder and Formatter for this Stringbuilder
            Formatter f = new Formatter(new StringBuilder());
            // get contents of lateFeeTable
            String[] lft = (readFile( clientLateFeeTableFileFullPath())).split(splitRegEx);
            // read total points from late fee table
            if (lft.length >= 2) {
                // append total score line to grade report
                f.format(formatString, Double.parseDouble(lft[0]), lft[1]);
                // initialize score accumulator
                total = Double.parseDouble(lft[0]);
            }
            // read points from grade file
            String[] gf = (readFile( clientGradeFileFullPath())).split(splitRegEx);
            // append grades to the report
            for (int i = 0; i < gf.length; i = i + 2) {
                d = Double.parseDouble(gf[i]);
                f.format(formatString, d, gf[i + 1]);
                // accumulate score
                total += d;
            }
            // append any late fees
            // get current date/time
            Date now = new Date();
            // loop over dates in LateFeeTable
            for (int i = 2; i < lft.length; i = i + 2) {
                // if date read is before now, add late fee
                if ((new SimpleDateFormat(DATEFORMATSTRING)).parse(lft[i+1],
                        new ParsePosition(0)).compareTo(now) < 0) {
                    d = Double.parseDouble(lft[i]);
                    f.format(formatString, d, "Received after " + lft[i+1]);
                    total += d;
                } else // date read is equal or after now
                    break;
            }
            
            // add total to report
            f.format(formatString, total, "Total Score");
            // delete grade file
			(new File(clientGradeFileFullPath())).delete();
            // add summar line to top of report for grade file
            String gfString = userID + "," + labNo + "," +
				total + "\n" + f.toString();
            // rewrite grade file
            writeFile(clientGradeFileFullPath(), gfString);
            // copy gradefile to gui
            gui.appendGrade(f.toString());

        } catch (Exception ex) {
            showMsg("EXCEPTION: UNABLE TO POST GRADE:\n");
            redirectStdIO();
            ex.printStackTrace();
            showMsg(baos.toString());
        }
    }

    /**
     * @param a file name without any path 
     * @return a string that is the full path name
     */
    public static String getClientFileFullPath(String fileName)
    {
        return userDir + File.separator + fileName;
    }

    /**
     * get the program under test and the latefee table
     * @return true if got some have something we can use
     */
    public static boolean getServerDataFiles()
    {
        boolean retVal = false; // assume the worst
        try {
            showMsg("RECEIVING Test Harness files ...\n");
            redirectStdIO();
            // Get PUT
            getFromServer(
				courseNo + "/" + labNo + "/" + HARNESSZIPFILENAME,
				userDir + File.separator);

            // got it
            // unzip it.
            retVal = unZip(getClientFileFullPath(HARNESSZIPFILENAME), userDir);
        } catch (Exception e) {
            try {
				// couldn't get file from server, 
				// try to get local copy of test harness
				retVal = (new File(
					getClientFileFullPath(HARNESSZIPFILENAME))).canRead() &&
					unZip( getClientFileFullPath(HARNESSZIPFILENAME),
						userDir);
				// if found a usable local copy 
				if ( retVal) {
					showMsg( "COULD NOT GET Test Harness files FROM SERVER " +
						" -- USING LOCAL COPY.\n");
					usingLocalTestHarness = true;
				}
            } catch (Exception ex) {
                showMsg("\nEXCEPTION when trying to receive Test Harness files:\n");
                ex.printStackTrace();
                System.err.println();
                e.printStackTrace();
            }
        } finally {
            showMsg(baos.toString());
            return retVal;
        }
    }

    /**
     * Uncompress files in a zip file to a directory
     * @param zipFileName Name of the zip file
     * @param recDir Full path to the directory to receive the files
     * @return true if successful, false otheriwse -- catches all exceptions
     */
    private static boolean unZip(String zipFileName, String recDir) 
    {
        try {
            // Create a buffer for reading the files
            byte[] buf = new byte[2048];

            // Create the ZIP input stream
            ZipInputStream zin = new ZipInputStream( new FileInputStream(zipFileName));

            // init variable to hold entries in the zip file
            ZipEntry ze = null;

            // UnCompress the files:
            // While there are entries
            while ((ze = zin.getNextEntry()) != null) {
                // Transfer bytes for this entry:
                // Create a file with the name of this entry
                FileOutputStream out = new FileOutputStream(
                    recDir + File.separator + ze.getName());
                int len = 0;
                // As long as bytes are available for this entry
                while (zin.available() > 0) {
                    // read bytes
                    len = zin.read(buf, 0, 2048);
                    // if there were bytes to read
                    if (len > 0)
                        // write bytes out to the file
                        out.write(buf, 0, len);
                }
                // Complete the entry:
                // close the output file
                out.close();
                // close the entry
                zin.closeEntry();
            }
            // close the zip file
            zin.close();
            return true;
        } catch (Exception ex) {
            showMsg("\nEXCEPTION when trying to unzip files from " +
				zipFileName + "\n");
            ex.printStackTrace();
            System.err.println();
            return false;
        }
            
    }

    /**
     * Creates a new byte output stream and 
     * redirects stdout, and err to this stream
     */
    private static void redirectStdIO()
    {
         baos = new ByteArrayOutputStream();
         // redirect System.out and System.err to the new baos
         System.setOut(new PrintStream(baos));
         System.setErr(new PrintStream(baos));
    }

    /**
     * appends a message to the log.
     */
    public static void showMsg(String s)
    {
        gui.appendLog(s);
    }

    /**
     * Downloads a file to the client
     * Optionally saves file as a string
	 * opens a new connection and closes the connection when finished.
     * @param sourceName name of server file -- relative to
     * root of autograder's directory structure on server.
     * @param targetDirName Directory name in which to write local file
     * false
     */
    private static void getFromServer(
		String sourceName, String targetDirName)
        throws Exception
    {
		// connect to server
		SCP scp = new SCP();
		// do transfer
		scp.transfer( SCP.GET, sourceName, targetDirName);
		// close connection
		scp.close();
    }

    /**
     * Uploads a file to the server
     * @param sourceName file or dir name on local machine
     * @param targetDirName directory name on server
     */
    private static void putToServer(String sourceName, String targetDirName)
        throws Exception
    {
		// connect to server
		SCP scp = new SCP();
		// do transfer
		scp.transfer( SCP.PUT, sourceName, targetDirName);
		// close connection
		scp.close();
    }

    /**
     * Read a file into a string
     * @param fName the file name
     * @return a string containing the bytes in the file or the empty string
     */
    public static String readFile( String fName) throws Exception
    {
        String retVal = "";
        InputStream is = new FileInputStream(new File(fName));
        // Create the byte array to hold the data
        byte[] bytes = new byte[2048];
    
        // Read in the bytes
        // keep reading as long as bytes are available
        while (is.available() > 0) {
            // read the bytes
            int numRead = is.read(bytes, 0, 2048);
            // append to the string the bytes read
            retVal += new String(bytes, 0, numRead);
        }
    
        // Close the input stream and return bytes
        is.close();
        return retVal;
    }

    /**
     * Write a file from a string
     * @param fName the file name to write
     * @param text the content to write to the file
     */
    public static void writeFile( String fName, String text) throws Exception
    {
        FileOutputStream f = new FileOutputStream(fName);
        DataOutputStream dos;
        dos = new DataOutputStream(new BufferedOutputStream(f,128));
        dos.writeBytes(text);
        dos.close();
        f.close();
    }

    /**
     * get files in dir whose names end with endsWith
     * @param dir the name of the directory to search
     * @param endsWith the file extension (end of the file name)
     * @return array of files -- empty if no files match or dirName does not name
     * a directory
     */
    private static File[] getFilesInDirEndingWith(String dirName, String endsWith)
    {
        File[] retVal = (new File(dirName)).listFiles();
        if (retVal == null)
            retVal = new File[0];
        else { // there are files in the list
            // filter out the ones I don't want
            Vector v = new Vector();
            for (int i = 0; i < retVal.length; i++)
                if (retVal[i].getName().endsWith(endsWith))
                    v.add(retVal[i]);
            retVal = new File[v.size()];
            for (int i = 0; i < v.size(); i++)
                retVal[i] = (File) v.elementAt(i);
        }
        return retVal;
    }

    /**
     * Recursively list the files and directories in a directory
     * from http://javaalmanac.com/egs/java.io/GetFiles.html and
     * http://javaalmanac.com/egs/java.io/TraverseTree.html?l=rel
     * @param dir The directory to list
     * @param fileNameEndsWith if non-null, files names must end with this string 
     * @return array containing the list
     */
    private static String[] recursiveDirList(File dir, String fileNameEndsWith)
    {
        // create return array with 0 elements
        String[] retArr = new String[0];
        // if I am a file
        if (dir.isFile()) {
            // return me if not checking for matches or are checking and I match 
            if (fileNameEndsWith == null || dir.getName().endsWith(fileNameEndsWith))
                retArr =  new String[]{dir.getName()};
        }
        else { // I am a directory
            // get my children
            String[] children = dir.list();
            // for each child
            for (int i=0; i<children.length; i++) {
                // get this child's list
                String[] childList =
                    recursiveDirList(new File(dir, children[i]), fileNameEndsWith);
                // create an array to add this child's list to the current total
                String[] tmpArr = new String[retArr.length + childList.length];
                // transfer existing elements to the temp array
                int j;
                for (j=0; j<retArr.length; j++)
                    tmpArr[j] = retArr[j];
                // transfer child's list to the temp array
                for (int k = 0; k<childList.length; k++)
                    tmpArr[j++] = dir.getName() + File.separator + childList[k];
                // save total array
                retArr = tmpArr;
            }
        }
        return retArr;
    }

    /**
     * @return the full path of the grade file on the client machine
     */
    public static String clientGradeFileFullPath()
    {
        return getClientFileFullPath(userID + "." + labNo + GRADEFILEEXT);
    }

    /**
     * @return the full path of the LateFeeTable file on the client machine
     */
    public static String clientLateFeeTableFileFullPath()
    {
        return getClientFileFullPath(LATEFEETABLENAME);
    }

    /**
     * @return the full path of the zip file on the client machine
     */
    public static String clientZipFileFullPath()
    {
        return getClientFileFullPath(userID + "." + labNo + ZIPEXT);
    }

    /**
     * Make a zip of all files in the user's directory
     */
    private static void createZipFile() throws Exception
    {
        showMsg("CREATING ZIP FILE ...\n");
        // create the name of the zip file
        String outFilename = clientZipFileFullPath();
        // delete any existing zip file for this lab
        (new File(outFilename)).delete();

        // make a file for userDir
        File userDirFile = new File(userDir);
        // list the userDir directory
        String[] filenames = recursiveDirList(userDirFile, null);

        // prepend userDir to each pathname
        String userDirParent = userDirFile.getParent() + File.separator;

        //javaalmanac.com/egs/java.util.zip/CreateZip.html
        // from: http://javaalmanac.com/egs/java.util.zip/CreateZip.html

        // Create a buffer for reading the files
        byte[] buf = new byte[2048];

        // Create the ZIP file
        ZipOutputStream out = new ZipOutputStream(new FileOutputStream(outFilename));

        // Compress the files
        for (int i=0; i<filenames.length; i++) {
            FileInputStream in =
                new FileInputStream(userDirParent + File.separator + filenames[i]);

            // Add ZIP entry to output stream.
            out.putNextEntry(new ZipEntry(filenames[i]));

            // Transfer bytes from the file to the ZIP file
            int len;
            while ((len = in.read(buf)) > 0) {
                out.write(buf, 0, len);
            }

            // Complete the entry
            out.closeEntry();
            in.close();
        }

        // Complete the ZIP file
        out.close();
    }

    /**
     * Gather all results and send them up to the server
     * Logs all exceptions to the gui -- gets as many as can to the log file
     */
    private static void sendResults()
    {
        // get a clean log
        redirectStdIO();
        // assume the worst
        boolean attemptShipToServer = false;
        try {
            // write out the log file
            writeFile(getClientFileFullPath(LOGFILENAME), gui.getLog());
            // bundle all files for shipment to server
            createZipFile();
            attemptShipToServer = true;
        } catch (Exception ex) {
            showMsg("\nUNABLE TO WRITE LOCAL RESULT FILES -- \n");
            ex.printStackTrace();
        } finally {
            //
            try {
                // if able to write local files
                if (attemptShipToServer) {
                    // try to ship to server
                    showMsg("\nSENDING RESULTS TO SERVER ...\n");
                    shipResults();
                }
            } catch (Exception ex) {
                showMsg("\nUNABLE TO SEND RESULT FILES TO SERVER:\n" + ex.toString());
                ex.printStackTrace();
            } finally {
                // flush the output buffer -- user display will show all errors
                showMsg(baos.toString());
            }
        }
    }

    /**
     * sends the zip file and the grade file up to the server
     **/
    private static void shipResults() throws Exception
    {
		// connect to server
		SCP scp = new SCP();

        // send grade file
		scp.transfer(
			SCP.PUT,
            clientGradeFileFullPath(),
            courseNo + "/" + labNo + "/" + GRADEDIRNAME);
        // send zip file
		scp.transfer(
			SCP.PUT,
            clientZipFileFullPath(),
            courseNo + "/" + labNo + "/" + GRADEDIRNAME);
		// close connection
		scp.close();
    }
}
