import javax.swing.SpringLayout;
import javax.swing.Spring;
import javax.swing.JOptionPane;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
    AutoGrader - a program that automatically grades programming assignments.
    Copyright (C) 2006  Pete Nordquist   (nordquip at sou.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
	(See the file COPYING)


	This program uses code that is Copyright (c) 2005 - 2006 Swiss Federal
	Institute of Technology (ETH Zurich),
	Department of Computer Science (http://www.inf.ethz.ch),
	Christian Plattner. All rights reserved.
	(See the file LICENSE.txt)

 * GUI for the AG application.
 * @author Pete Nordquist
 * @version 061010
 * @version 070928 - corrected components relationship, which fixed
 * course list window extending log window bug.
 */

public class AGGUI extends JFrame implements ActionListener
{
    // constants
	private static final String LS = System.getProperty("line.separator");
    private static final int NUMUSERIDCOLS = 8;
    private static final int NUMLABNOROWS = 6;
    private static final int NUMLOGROWS = 15;
    private static final int NUMLOGCOLS = 80;
    private static final String LOGFONTNAME = "Monospaced";
    private static final int LOGFONTSTYLE = Font.PLAIN;
    private static final int LOGFONTSIZE = 12;
    private static final int NUMGRADEROWS = 15;
    private static final int NUMGRADECOLS = 80;
    private static final String GRADEFONTNAME = "Monospaced";
    private static final int GRADEFONTSTYLE = Font.PLAIN;
    private static final int GRADEFONTSIZE = 12;
    // courses to appear in the course list
    private static final String [] COURSEARRAY = {"cs257", "cs200"};
    // GUI actions that call the PUT application
    private static final String SUBMIT = "Submit";

    // create labels
    private JLabel courseNoLabel = new JLabel("Select course:");
    private JLabel userDirLabel = new JLabel("Enter full path to the directory " +
        "containing your .java files, e.g.:  F:\\cs257\\lab1");

    // make labels for the input areas
    // need a wide label for User ID, so lab no will be far enough to the right
    private JLabel userIDLabel = new JLabel("Enter User ID:        ");
    private JLabel labNoLabel = new JLabel("Select Lab Number:");
	private String defaultToolsJarPath = 
		"C:\\Program Files\\Java\\jdk1.6.0_07\\lib";
    private JLabel toolsJarPathLabel = new JLabel("Enter full path to local " +
        "tools.jar file directory, e.g.:  " +
		defaultToolsJarPath);

    // log label
    private JLabel logLabel = new JLabel("OUTPUT LOG:");

    // grade label
    private JLabel gradeLabel = new JLabel("TEST RESULTS:");

    // input fields
    private TextField userID = new TextField(NUMUSERIDCOLS);
    // userIDList -- not displayed -- used for checking the userID
    private String[] userIDList = null;
    private List labNo = new List(NUMLABNOROWS, false);
    private List courseNo = new List(COURSEARRAY.length, false);
    private TextField userDir = new TextField(NUMLOGCOLS);
    private TextField toolsJarPath = new TextField(NUMLOGCOLS);
    // submit for page 1
    private JButton submit1 = new JButton("Submit");
    // submit for page 2
    private JButton submit2 = new JButton("Submit");
    // request Key button
    private JButton sendKey = new JButton("Get Instructor's Solution");

    // log area
    private TextArea log = new TextArea(NUMLOGROWS, NUMLOGCOLS);

    // grade display
    private TextArea grade = new TextArea(NUMGRADEROWS, NUMGRADECOLS);

    /****************************
     P U B L I C   M E T H O D S
    ****************************/

    /**
     * Create a new AG window.
     */
    public AGGUI() {
        super("SOU Automatic Grader for Java Labs");

        // initialize components
        initList(courseNo, COURSEARRAY);
		toolsJarPath.setText(defaultToolsJarPath);

        // log area
        log.setFont(new Font(LOGFONTNAME, LOGFONTSTYLE, LOGFONTSIZE));

        // grade field
        grade.setFont(new Font(GRADEFONTNAME, GRADEFONTSTYLE, GRADEFONTSIZE));

        // default close operation
        this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        displayPage1();
    }

    /****************************************************
        Accessors for the data
    ****************************************************/
    public void appendLog(String s)
    {
        log.append(s);
    }
    public String getLog()
    {
        return log.getText();
    }
    public String getUserID()
    {
        return userID.getText();
    }
    public String getLabNo()
    {
        return labNo.getSelectedItem();
    }
    public String getCourseNo()
    {
        return courseNo.getSelectedItem();
    }
    public String getUserDir()
    {
        return userDir.getText();
    }
    public String getToolsJarPath()
    {
        return toolsJarPath.getText();
    }
    public void appendGrade(String s)
    {
        grade.append(s);
    }

    /**
     * remove all fields except the data area
     */
    public void removeAllButLog()
    {
        // make it invisible
        this.setVisible(false);

        //Set up the content pane.
        Container contentPane = this.getContentPane();
        // remove page 1 components
        contentPane.removeAll();
        // create a new layout manager --
		// may have been trouble reusing the old one
        SpringLayout layout = new SpringLayout();
        contentPane.setLayout(layout);

        // build up page 2
        // add components to contentPane
        contentPane.add(logLabel);
        contentPane.add(log);

		// NW of logLabel relative to SW of submit2
		layout.putConstraint(SpringLayout.WEST,
			logLabel,
			5,
			SpringLayout.WEST,
			contentPane);
		layout.putConstraint(SpringLayout.NORTH,
			logLabel,
			5,
			SpringLayout.NORTH,
			contentPane);

		// NW of log relative to SW of logLabel
		layout.putConstraint(SpringLayout.WEST,
			log,
			0,
			SpringLayout.WEST,
			logLabel);
		layout.putConstraint(SpringLayout.NORTH,
			log,
			5,
			SpringLayout.SOUTH,
			logLabel);

		// give the contentPane total dimensions
		layout.putConstraint(SpringLayout.EAST,
			contentPane,
			5,
			SpringLayout.EAST,
			log);
		layout.putConstraint(SpringLayout.SOUTH,
			contentPane,
			5,
			SpringLayout.SOUTH,
			log);


        makeItVisible();
    }

    /**
     * listener that dispatches actions
     */
    public void actionPerformed(ActionEvent event)
    {
        // if page1 submit, get 
        if (event.getSource() == submit1)
        {
            // if no course selected
            if (this.getCourseNo() == null) 
                appendLog(LS+"?NO COURSE IS SELECTED -- please try again."+LS);

            else {
                // walk the list of possible courses
                for (int i = 0; i < COURSEARRAY.length; i++)
                {
                    // if selected course matches
                    if (this.getCourseNo().equals(COURSEARRAY[i])) {
                        // check that text fields are OK
                        if (! AG.checkAndSaveGUIPaths()) 
                            appendLog("  -- please try again."+LS);
                        else {
                            // get userID and labNo lists for this course
                            Object[] oa = AG.populateGUILists();
                            // if did not successfully get some lists
                            if (! ((Boolean) oa[2]).booleanValue())
                                AG.quit();
                            // got lists, so go on
                            else {
                                // save the userIDList for later checking
                                userIDList = (String[]) oa[0];
                                // populate the labNo list
                                this.initList(labNo, (String[]) oa[1]);
                                // display page 2
                                displayPage2();
                            } 
                        }
                        // quit the loop
                        break;
                    }
                }
            }
        }
        // if page2 submit, call AG
        else if (event.getSource() == submit2) {
            // if OK, then submit
            if (checkUserIDAndLabNo())
                AG.doSubmit();
        }
        // sendKey
        else if (event.getSource() == sendKey &&
            // check for legal ID and lab selected
            checkUserIDAndLabNo() &&
            // Confirm with a dialog box
            JOptionPane.showInputDialog("No submissions are allowed after a solution " +
                "is received."+LS+
				"Really get solution?").trim().toLowerCase().charAt(0) == 'y')
                AG.receiveKey();
    }

    /****************************
     P R I V A T E   M E T H O D S
    ****************************/
    /**
     * Check input data -- print errors if not valid
     * @return true if userID is valid and a labNo is selected.  false otherwise
     */
    private boolean checkUserIDAndLabNo()
    {
        // check userID
        // assume the worst
        boolean userIDOK = false;
        for (int i = 0; i < userIDList.length; i++) {
            if (getUserID().equals(userIDList[i])) {
                userIDOK = true;
                break;
            }
        }
        if (! userIDOK)
            appendLog("?USERID DOES NOT MATCH -- please try again."+LS);

        // check that some labNo is selected
        boolean labNoOK = false;
        if ( getLabNo() != null)
            labNoOK = true;
        else
            appendLog("?NO LAB NUMBER IS SELECTED -- please try again."+LS);

        return userIDOK && labNoOK;
    }

    /**
     * Displays the first page of the GUI:
     * courseNo
     * userDir
     * submit
     * log
     */
    private void displayPage1()
    {

        //Set up the content pane.
        Container contentPane = this.getContentPane();
        SpringLayout layout = new SpringLayout();
        contentPane.setLayout(layout);

        // add components to contentPane
        contentPane.add(courseNoLabel);
        contentPane.add(courseNo);
        contentPane.add(userDirLabel);
        contentPane.add(userDir);
        contentPane.add(toolsJarPathLabel);
        contentPane.add(toolsJarPath);
        contentPane.add(submit1);
        contentPane.add(logLabel);
        contentPane.add(log);
		
		// set sizes for fields that shouldn't grow by setting the max
		// size to the original preferred size
		// when max size is same as preferred size, SpringLayout does not
		// try to stretch the component
		courseNo.setMaximumSize(courseNo.getPreferredSize());
		userDir.setMaximumSize(userDir.getPreferredSize());
		toolsJarPath.setMaximumSize(toolsJarPath.getPreferredSize());

		// NW of courseNoLabel relative to NW of contentPane
		layout.putConstraint(SpringLayout.WEST,
			courseNoLabel,
			5,
			SpringLayout.WEST,
			contentPane);
		layout.putConstraint(SpringLayout.NORTH,
			courseNoLabel,
			5,
			SpringLayout.NORTH,
			contentPane);

		// NW of courseNO relative to NE of courseNoLabel
		layout.putConstraint(SpringLayout.WEST,
			courseNo,
			5,
			SpringLayout.EAST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			courseNo,
			0,
			SpringLayout.NORTH,
			courseNoLabel);

		// must make components relative to each other, not the contentPane
		// or they will not hang together when the window resizes
		// NW of userDirLabel, below courseNo
		layout.putConstraint(SpringLayout.WEST,
			userDirLabel,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			userDirLabel,
			5,
			SpringLayout.SOUTH,
			courseNo);

		// NW of userDir edge below userDirLabel
		layout.putConstraint(SpringLayout.WEST,
			userDir,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			userDir,
			5,
			SpringLayout.SOUTH,
			userDirLabel);

		// NW of toolsJarPathLabel below userDir
		layout.putConstraint(SpringLayout.WEST,
			toolsJarPathLabel,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			toolsJarPathLabel,
			5,
			SpringLayout.SOUTH,
			userDir);

		// NW of toolsJarPath below toolsJarPathLabel
		layout.putConstraint(SpringLayout.WEST,
			toolsJarPath,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			toolsJarPath,
			5,
			SpringLayout.SOUTH,
			toolsJarPathLabel);

		// toolsJarPath same width as userDir
		layout.putConstraint(SpringLayout.EAST,
			toolsJarPath,
			0,
			SpringLayout.EAST,
			userDir);

		// NW of submit1 below toolsJarPath
		layout.putConstraint(SpringLayout.WEST,
			submit1,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			submit1,
			5,
			SpringLayout.SOUTH,
			toolsJarPath);

		// NW of logLabel below submit1
		layout.putConstraint(SpringLayout.WEST,
			logLabel,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			logLabel,
			5,
			SpringLayout.SOUTH,
			submit1);

		// NW of log below logLabel
		layout.putConstraint(SpringLayout.WEST,
			log,
			0,
			SpringLayout.WEST,
			courseNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			log,
			5,
			SpringLayout.SOUTH,
			logLabel);

		// give the contentPane total dimensions
		layout.putConstraint(SpringLayout.EAST,
			contentPane,
			5,
			SpringLayout.EAST,
			log);
		layout.putConstraint(SpringLayout.SOUTH,
			contentPane,
			5,
			SpringLayout.SOUTH,
			log);

        // create responder
        submit1.addActionListener(this);

        //Display the window.
        makeItVisible();
    }
    
    private void makeItVisible()
    {
        this.setBackground(SystemColor.control);
        this.pack();
        this.setVisible(true);
    }

    /**
     *  Displays page 2:
     *  userDir selected
     *  userID    LabNo
     *  toolsJarPath
     *  submit
     *  log
     *  grade
     */
    private void displayPage2()
    {
        // make it invisible
        this.setVisible(false);

        //Set up the content pane.
        Container contentPane = this.getContentPane();
        // remove page 1 components
        contentPane.removeAll();
        // create a new layout manager --
		// may have been trouble reusing the old one
        SpringLayout layout = new SpringLayout();
        contentPane.setLayout(layout);

        // build up page 2
        // add components to contentPane
        contentPane.add(userIDLabel);
        contentPane.add(userID);
        contentPane.add(labNoLabel);
        contentPane.add(labNo);
        contentPane.add(sendKey);
        contentPane.add(submit2);
        contentPane.add(logLabel);
        contentPane.add(log);
        contentPane.add(gradeLabel);
        contentPane.add(grade);

		// set sizes for fields that shouldn't grow by setting the max
		// size to the original preferred size
		// when max size is same as preferred size, SpringLayout does not
		// try to stretch the component
		userID.setMaximumSize(userID.getPreferredSize());
		labNo.setMaximumSize(labNo.getPreferredSize());

		// NW of userIDLabel relative to NW of contentPane
		layout.putConstraint(SpringLayout.WEST,
			userIDLabel,
			5,
			SpringLayout.WEST,
			contentPane);
		layout.putConstraint(SpringLayout.NORTH,
			userIDLabel,
			0,
			SpringLayout.NORTH,
			labNo);

		// NW of labNoLabel relative to N of userIDLabel and E of userID
		layout.putConstraint(SpringLayout.WEST,
			labNoLabel,
			35,
			SpringLayout.EAST,
			userID);
		layout.putConstraint(SpringLayout.NORTH,
			labNoLabel,
			0,
			SpringLayout.NORTH,
			labNo);

		// NW of labNo relative to N of userIDLabel and E of labNoLabel
		layout.putConstraint(SpringLayout.WEST,
			labNo,
			5,
			SpringLayout.EAST,
			labNoLabel);
		layout.putConstraint(SpringLayout.NORTH,
			labNo,
			5,
			SpringLayout.NORTH,
			contentPane);

		// NW of sendKey relative to N of userIDLabel and E of labNo
		layout.putConstraint(SpringLayout.WEST,
			sendKey,
			5,
			SpringLayout.EAST,
			labNo);
		layout.putConstraint(SpringLayout.VERTICAL_CENTER,
			sendKey,
			0,
			SpringLayout.VERTICAL_CENTER,
			labNo);

		// NW of userID relative to SW of userIDLabel
		layout.putConstraint(SpringLayout.WEST,
			userID,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			userID,
			5,
			SpringLayout.SOUTH,
			userIDLabel);

		// NW of submit2 relative to SW of userID
		layout.putConstraint(SpringLayout.WEST,
			submit2,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			submit2,
			5,
			SpringLayout.SOUTH,
			userID);

		// NW of logLabel relative to SW of submit2
		layout.putConstraint(SpringLayout.WEST,
			logLabel,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			logLabel,
			5,
			SpringLayout.SOUTH,
			submit2);

		// NW of log relative to SW of logLabel
		layout.putConstraint(SpringLayout.WEST,
			log,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			log,
			5,
			SpringLayout.SOUTH,
			logLabel);

		// NW of gradeLabel relative to SW of log
		layout.putConstraint(SpringLayout.WEST,
			gradeLabel,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			gradeLabel,
			5,
			SpringLayout.SOUTH,
			log);

		// NW of grade relative to SW of gradeLabel
		layout.putConstraint(SpringLayout.WEST,
			grade,
			0,
			SpringLayout.WEST,
			userIDLabel);
		layout.putConstraint(SpringLayout.NORTH,
			grade,
			5,
			SpringLayout.SOUTH,
			gradeLabel);

		// width of grade same as width of log
		layout.putConstraint(SpringLayout.EAST,
			grade,
			0,
			SpringLayout.EAST,
			log);

		// give the contentPane total dimensions
		layout.putConstraint(SpringLayout.EAST,
			contentPane,
			5,
			SpringLayout.EAST,
			log);
		layout.putConstraint(SpringLayout.SOUTH,
			contentPane,
			5,
			SpringLayout.SOUTH,
			grade);

        // create responders
        submit2.addActionListener(this);
        sendKey.addActionListener(this);

        makeItVisible();
    }

    /**
     * Initialize a list with an array of strings
     * @param l the list
     * @param sa the strings to put in the list
     */
    private void initList(List l, String[] sa)
    {
        for (int i = 0; i < sa.length; i++)
            l.add(sa[i]);
    }
}
